"""Tests for the event store."""

# pylint: disable=too-many-lines

from __future__ import annotations

import datetime
import zoneinfo
from collections.abc import Callable, Generator
from typing import Any
from unittest.mock import patch

import pytest
from freezegun import freeze_time
from freezegun.api import FrozenDateTimeFactory

from ical.calendar import Calendar
from ical.event import Event
from ical.store import EventStore, EventStoreError
from ical.types.recur import Range, Recur


@pytest.fixture(name="calendar")
def mock_calendar() -> Calendar:
    """Fixture to create a calendar."""
    return Calendar()


@pytest.fixture(name="store")
def mock_store(calendar: Calendar) -> EventStore:
    """Fixture to create an event store."""
    return EventStore(calendar)


@pytest.fixture(name="_uid", autouse=True)
def mock_uid() -> Generator[None, None, None]:
    """Patch out uuid creation with a fixed value."""
    counter = 0

    def func() -> str:
        nonlocal counter
        counter += 1
        return f"mock-uid-{counter}"

    with patch("ical.event.uid_factory", new=func):
        yield


@pytest.fixture(name="fetch_events")
def mock_fetch_events(
    calendar: Calendar,
) -> Callable[..., list[dict[str, Any]]]:
    """Fixture to return events on the calendar."""

    def _func(keys: set[str] | None = None) -> list[dict[str, Any]]:
        result = []
        for event in calendar.timeline:
            data = event.dict()
            for key, value in list(data.items()):
                if value is None or isinstance(value, list) and not value:
                    del data[key]
                elif keys and key not in keys:
                    del data[key]
                elif isinstance(value, datetime.datetime):
                    data[key] = value.isoformat()
                elif isinstance(value, datetime.date):
                    data[key] = value.isoformat()
            result.append(data)
        return result

    return _func


@pytest.fixture(name="frozen_time", autouse=True)
def mock_frozen_time() -> Generator[FrozenDateTimeFactory, None, None]:
    """Fixture to freeze time to a specific point."""
    with freeze_time("2022-09-03T09:38:05") as freeze:
        with patch("ical.event.dtstamp_factory", new=freeze):
            yield freeze


def test_empty_store(fetch_events: Callable[..., list[dict[str, Any]]]) -> None:
    """Test iteration over an empty calendar."""
    assert fetch_events() == []


def test_add_and_delete_event(
    store: EventStore, fetch_events: Callable[..., list[dict[str, Any]]]
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
        )
    )
    assert fetch_events() == [
        {
            "dtstamp": "2022-09-03T09:38:05",
            "uid": "mock-uid-1",
            "created": "2022-09-03T09:38:05",
            "dtstart": "2022-08-29T09:00:00",
            "dtend": "2022-08-29T09:30:00",
            "summary": "Monday meeting",
            "sequence": 0,
        },
    ]
    store.delete("mock-uid-1")
    assert fetch_events() == []


def test_edit_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
        )
    )
    assert fetch_events() == [
        {
            "dtstamp": "2022-09-03T09:38:05",
            "uid": "mock-uid-1",
            "created": "2022-09-03T09:38:05",
            "dtstart": "2022-08-29T09:00:00",
            "dtend": "2022-08-29T09:30:00",
            "summary": "Monday meeting",
            "sequence": 0,
        },
    ]

    frozen_time.tick(delta=datetime.timedelta(seconds=10))

    # Set event start time 5 minutes later
    store.edit(
        "mock-uid-1",
        Event(start="2022-08-29T09:05:00", summary="Monday meeting (Delayed)"),
    )
    assert fetch_events() == [
        {
            "dtstamp": "2022-09-03T09:38:15",
            "uid": "mock-uid-1",
            "created": "2022-09-03T09:38:05",
            "dtstart": "2022-08-29T09:05:00",
            "dtend": "2022-08-29T09:30:00",
            "summary": "Monday meeting (Delayed)",
            "sequence": 1,
            "last_modified": "2022-09-03T09:38:15",
        },
    ]


def test_edit_event_invalid_uid(store: EventStore) -> None:
    """Edit an event that does not exist."""
    with pytest.raises(EventStoreError, match="No existing"):
        store.edit("mock-uid-1", Event(start="2022-08-29T09:05:00", summary="Delayed"))


@pytest.mark.parametrize(
    ("start", "end", "recur", "results"),
    [
        (
            datetime.datetime(2022, 8, 29, 9, 0),
            datetime.datetime(2022, 8, 29, 9, 30),
            Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926T090000"),
            [
                "2022-08-29T09:00:00",
                "2022-09-05T09:00:00",
                "2022-09-12T09:00:00",
                "2022-09-19T09:00:00",
                "2022-09-26T09:00:00",
            ],
        ),
        (
            datetime.datetime(2022, 8, 29, 9, 0),
            datetime.datetime(2022, 8, 29, 9, 30),
            Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
            [
                "2022-08-29T09:00:00",
                "2022-09-05T09:00:00",
                "2022-09-12T09:00:00",
                "2022-09-19T09:00:00",
                "2022-09-26T09:00:00",
            ],
        ),
        (
            datetime.datetime(
                2022, 8, 29, 9, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
            datetime.datetime(
                2022, 8, 29, 9, 30, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
            Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
            [
                "2022-08-29T09:00:00-07:00",
                "2022-09-05T09:00:00-07:00",
                "2022-09-12T09:00:00-07:00",
                "2022-09-19T09:00:00-07:00",
                "2022-09-26T09:00:00-07:00",
            ],
        ),
    ],
)
def test_recurring_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    start: datetime.datetime,
    end: datetime.datetime,
    recur: Recur,
    results: list[str],
) -> None:
    """Test adding a recurring event and deleting the entire series."""
    store.add(
        Event(
            summary="Monday meeting",
            start=start,
            end=end,
            rrule=recur,
        )
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": results[0],
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220905T090000",
            "dtstart": results[1],
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220912T090000",
            "dtstart": results[2],
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220919T090000",
            "dtstart": results[3],
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220926T090000",
            "dtstart": results[4],
            "summary": "Monday meeting",
        },
    ]
    store.delete("mock-uid-1")
    assert fetch_events(None) == []


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
    ],
)
def test_deletel_partial_recurring_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    recur: Recur,
) -> None:
    """Test adding a recurring event and deleting part of the series."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    store.delete(uid="mock-uid-1", recurrence_id="20220905T090000")
    store.delete(uid="mock-uid-1", recurrence_id="20220919T090000")
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220912T090000",
            "dtstart": "2022-09-12T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220926T090000",
            "dtstart": "2022-09-26T09:00:00",
            "summary": "Monday meeting",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
    ],
)
def test_delete_this_and_future_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    recur: Recur,
) -> None:
    """Test adding a recurring event and deleting events after one event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    store.delete(
        uid="mock-uid-1",
        recurrence_id="20220919T090000",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220905T090000",
            "dtstart": "2022-09-05T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220912T090000",
            "dtstart": "2022-09-12T09:00:00",
            "summary": "Monday meeting",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
    ],
)
def test_delete_this_and_future_all_day_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    recur: Recur,
) -> None:
    """Test adding a recurring event and deleting events after one event."""
    store.add(
        Event(
            summary="Mondays",
            start="2022-08-29",
            end="2022-08-30",
            rrule=recur,
        )
    )
    store.delete(
        uid="mock-uid-1",
        recurrence_id="20220919",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829",
            "dtstart": "2022-08-29",
            "summary": "Mondays",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220905",
            "dtstart": "2022-09-05",
            "summary": "Mondays",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220912",
            "dtstart": "2022-09-12",
            "summary": "Mondays",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
    ],
)
def test_delete_this_and_future_event_with_first_instance(
    calendar: Calendar,
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    recur: Recur,
) -> None:
    """Test deleting this and future for the first instance."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    assert len(calendar.events) == 1
    store.delete(
        uid="mock-uid-1",
        recurrence_id="20220829T090000",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == []
    assert len(calendar.events) == 0


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220926"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=5"),
    ],
)
def test_delete_this_and_future_all_day_event_with_first_instance(
    calendar: Calendar,
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    recur: Recur,
) -> None:
    """Test deleting this and future for the first instance."""
    store.add(
        Event(
            summary="Mondays",
            start="2022-08-29",
            end="2022-08-29",
            rrule=recur,
        )
    )
    assert len(calendar.events) == 1
    store.delete(
        uid="mock-uid-1",
        recurrence_id="20220829",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == []
    assert len(calendar.events) == 0


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220913T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
    ],
)
def test_edit_recurring_event(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
    recur: Recur,
) -> None:
    """Test editing all instances of a recurring event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(start="2022-08-30T09:00:00", summary="Tuesday meeting"),
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220830T090000",
            "dtstart": "2022-08-30T09:00:00",
            "summary": "Tuesday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220906T090000",
            "dtstart": "2022-09-06T09:00:00",
            "summary": "Tuesday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220913T090000",
            "dtstart": "2022-09-13T09:00:00",
            "summary": "Tuesday meeting",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220912T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
    ],
)
def test_edit_recurring_event_instance(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
    recur: Recur,
) -> None:
    """Test editing all instances of a recurring event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(start="2022-09-06T09:00:00", summary="Tuesday meeting"),
        recurrence_id="20220905T090000",
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-2",
            "dtstart": "2022-09-06T09:00:00",
            "summary": "Tuesday meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220912T090000",
            "dtstart": "2022-09-12T09:00:00",
            "summary": "Monday meeting",
        },
    ]


def test_cant_change_recurrence_for_event_instance(
    store: EventStore,
    frozen_time: FrozenDateTimeFactory,
) -> None:
    """Test editing all instances of a recurring event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
        )
    )

    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    with pytest.raises(EventStoreError, match="single instance with rrule"):
        store.edit(
            "mock-uid-1",
            Event(
                start="2022-09-06T09:00:00",
                summary="Tuesday meeting",
                rrule=Recur.from_rrule("FREQ=DAILY;COUNT=3"),
            ),
            recurrence_id="20220905T090000",
        )


def test_convert_single_instance_to_recurring(
    store: EventStore,
    frozen_time: FrozenDateTimeFactory,
    fetch_events: Callable[..., list[dict[str, Any]]],
) -> None:
    """Test editing all instances of a recurring event."""
    store.add(
        Event(
            summary="Daily meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
        )
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Daily meeting",
        },
    ]

    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            summary="Daily meeting",
            rrule=Recur.from_rrule("FREQ=DAILY;COUNT=3"),
        ),
    )

    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Daily meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220830T090000",
            "dtstart": "2022-08-30T09:00:00",
            "summary": "Daily meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220831T090000",
            "dtstart": "2022-08-31T09:00:00",
            "summary": "Daily meeting",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220912T090000"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
    ],
)
def test_edit_recurring_event_this_and_future(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
    recur: Recur,
) -> None:
    """Test editing future instance of a recurring event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=recur,
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(summary="Team meeting"),
        recurrence_id="20220905T090000",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-2",
            "dtstart": "2022-09-05T09:00:00",
            "recurrence_id": "20220905T090000",
            "summary": "Team meeting",
        },
        {
            "uid": "mock-uid-2",
            "recurrence_id": "20220912T090000",
            "dtstart": "2022-09-12T09:00:00",
            "summary": "Team meeting",
        },
    ]


@pytest.mark.parametrize(
    "recur",
    [
        Recur.from_rrule("FREQ=WEEKLY;UNTIL=20220912"),
        Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
    ],
)
def test_edit_recurring_all_day_event_this_and_future(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
    recur: Recur,
) -> None:
    """Test editing future instance of a recurring event."""
    store.add(
        Event(
            summary="Monday",
            start="2022-08-29",
            end="2022-08-30",
            rrule=recur,
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(summary="Mondays [edit]"),
        recurrence_id="20220905",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829",
            "dtstart": "2022-08-29",
            "summary": "Monday",
        },
        {
            "uid": "mock-uid-2",
            "dtstart": "2022-09-05",
            "recurrence_id": "20220905",
            "summary": "Mondays [edit]",
        },
        {
            "uid": "mock-uid-2",
            "recurrence_id": "20220912",
            "dtstart": "2022-09-12",
            "summary": "Mondays [edit]",
        },
    ]


def test_add_and_delete_event_date(
    store: EventStore, fetch_events: Callable[..., list[dict[str, Any]]]
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29",
            end="2022-08-29",
        )
    )
    assert fetch_events() == [
        {
            "dtstamp": "2022-09-03T09:38:05",
            "uid": "mock-uid-1",
            "created": "2022-09-03T09:38:05",
            "dtstart": "2022-08-29",
            "dtend": "2022-08-29",
            "summary": "Monday meeting",
            "sequence": 0,
        },
    ]
    store.delete("mock-uid-1")
    assert fetch_events() == []


def test_delete_event_date_recurring(
    store: EventStore, fetch_events: Callable[..., list[dict[str, Any]]]
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29",
            end="2022-08-29",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
        )
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-08-29",
            "summary": "Monday meeting",
            "recurrence_id": "20220829",
        },
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-09-05",
            "summary": "Monday meeting",
            "recurrence_id": "20220905",
        },
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-09-12",
            "summary": "Monday meeting",
            "recurrence_id": "20220912",
        },
    ]

    store.delete("mock-uid-1", recurrence_id="20220905")
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-08-29",
            "summary": "Monday meeting",
            "recurrence_id": "20220829",
        },
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-09-12",
            "summary": "Monday meeting",
            "recurrence_id": "20220912",
        },
    ]


def test_edit_recurrence_rule_this_and_future(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
) -> None:
    """Test editing future instances of a recurring event."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(
            summary="Team meeting",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3;INTERVAL=2"),
        ),
        recurrence_id="20220905T090000",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220829T090000",
            "dtstart": "2022-08-29T09:00:00",
            "summary": "Monday meeting",
        },
        {
            "uid": "mock-uid-2",
            "dtstart": "2022-09-05T09:00:00",
            "recurrence_id": "20220905T090000",
            "summary": "Team meeting",
        },
        {
            "uid": "mock-uid-2",
            "recurrence_id": "20220919T090000",
            "dtstart": "2022-09-19T09:00:00",
            "summary": "Team meeting",
        },
    ]


def test_edit_recurrence_rule_this_and_future_all_day_first_instance(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
) -> None:
    """Test editing future instances starting at the first instance."""
    store.add(
        Event(
            summary="Monday",
            start="2022-08-29",
            end="2022-08-30",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(
            summary="Mondays [edit]",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3;INTERVAL=2"),
        ),
        recurrence_id="20220829",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-08-29",
            "recurrence_id": "20220829",
            "summary": "Mondays [edit]",
        },
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-09-12",
            "recurrence_id": "20220912",
            "summary": "Mondays [edit]",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220926",
            "dtstart": "2022-09-26",
            "summary": "Mondays [edit]",
        },
    ]


def test_edit_recurrence_rule_this_and_future_first_instance(
    store: EventStore,
    fetch_events: Callable[..., list[dict[str, Any]]],
    frozen_time: FrozenDateTimeFactory,
) -> None:
    """Test editing future instances starting at the first instance."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3"),
        )
    )
    frozen_time.tick(delta=datetime.timedelta(seconds=10))
    store.edit(
        "mock-uid-1",
        Event(
            summary="Team meeting",
            rrule=Recur.from_rrule("FREQ=WEEKLY;COUNT=3;INTERVAL=2"),
        ),
        recurrence_id="20220829T090000",
        recurrence_range=Range.THIS_AND_FUTURE,
    )
    assert fetch_events({"uid", "recurrence_id", "dtstart", "summary"}) == [
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-08-29T09:00:00",
            "recurrence_id": "20220829T090000",
            "summary": "Team meeting",
        },
        {
            "uid": "mock-uid-1",
            "dtstart": "2022-09-12T09:00:00",
            "recurrence_id": "20220912T090000",
            "summary": "Team meeting",
        },
        {
            "uid": "mock-uid-1",
            "recurrence_id": "20220926T090000",
            "dtstart": "2022-09-26T09:00:00",
            "summary": "Team meeting",
        },
    ]


def test_invalid_uid(
    store: EventStore,
) -> None:
    """Test iteration over an empty calendar."""
    with pytest.raises(EventStoreError, match=r"No existing event with uid"):
        store.edit("invalid", Event(summary="example summary"))

    with pytest.raises(EventStoreError, match=r"No existing event with uid"):
        store.delete("invalid")


def test_invalid_recurrence_id(
    store: EventStore,
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start="2022-08-29T09:00:00",
            end="2022-08-29T09:30:00",
        )
    )

    with pytest.raises(EventStoreError, match=r"event is not recurring"):
        store.delete("mock-uid-1", recurrence_id="invalid")

    with pytest.raises(EventStoreError, match=r"event is not recurring"):
        store.edit(
            "mock-uid-1", recurrence_id="invalid", event=Event(summary="invalid")
        )


def test_no_timezone_for_floating(
    calendar: Calendar,
    store: EventStore,
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start=datetime.datetime(2022, 8, 29, 9, 0, 0),
            end=datetime.datetime(2022, 8, 29, 9, 30, 0),
        )
    )
    assert len(calendar.events) == 1
    assert not calendar.timezones


def test_no_timezone_for_utc(
    calendar: Calendar,
    store: EventStore,
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start=datetime.datetime(2022, 8, 29, 9, 0, 0, tzinfo=datetime.timezone.utc),
            end=datetime.datetime(2022, 8, 29, 9, 30, 0, tzinfo=datetime.timezone.utc),
        )
    )
    assert len(calendar.events) == 1
    assert not calendar.timezones


def test_timezone_for_datetime(
    calendar: Calendar,
    store: EventStore,
) -> None:
    """Test adding an event to the store and retrieval."""
    store.add(
        Event(
            summary="Monday meeting",
            start=datetime.datetime(
                2022, 8, 29, 9, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
            end=datetime.datetime(
                2022, 8, 29, 9, 30, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
        )
    )
    assert len(calendar.events) == 1
    assert len(calendar.timezones) == 1
    assert calendar.timezones[0].tz_id == "America/Los_Angeles"

    store.add(
        Event(
            summary="Tuesday meeting",
            start=datetime.datetime(
                2022, 8, 30, 9, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
            end=datetime.datetime(
                2022, 8, 30, 9, 30, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
        )
    )
    # Timezone already exists
    assert len(calendar.timezones) == 1

    store.add(
        Event(
            summary="Wednesday meeting",
            start=datetime.datetime(
                2022, 8, 31, 12, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
            ),
            end=datetime.datetime(
                2022, 8, 31, 12, 30, 0, tzinfo=zoneinfo.ZoneInfo("America/New_York")
            ),
        )
    )
    assert len(calendar.timezones) == 2
    assert calendar.timezones[0].tz_id == "America/Los_Angeles"
    assert calendar.timezones[1].tz_id == "America/New_York"


def test_timezone_offset_not_supported(
    calendar: Calendar,
    store: EventStore,
) -> None:
    """Test adding a datetime for a timestamp that does not have a valid timezone."""
    offset = datetime.timedelta(hours=-8)
    tzinfo = datetime.timezone(offset=offset)
    event = Event(
        summary="Monday meeting",
        start=datetime.datetime(2022, 8, 29, 9, 0, 0, tzinfo=tzinfo),
        end=datetime.datetime(2022, 8, 29, 9, 30, 0, tzinfo=tzinfo),
    )
    with pytest.raises(EventStoreError, match=r"No timezone information"):
        store.add(event)
    assert not calendar.events
    assert not calendar.timezones



"""Library for handling rfc5545 properties and parameters.

A property is the definition of an individual attribute describing a
calendar object or a calendar component. A property is also really
just a "contentline", however properties in this file are the
output of the parser and are provided in the context of where
they live on a component hierarchy (e.g. attached to a component,
or sub component).
"""

# mypy: allow-any-generics

from __future__ import annotations

import datetime
from dataclasses import dataclass
from typing import Optional, Union

from .const import PARSE_PARAM_NAME, PARSE_PARAM_VALUE, PARSE_PARAMS
from .unicode import UNSAFE_CHAR_RE


@dataclass
class ParsedPropertyParameter:
    """An rfc5545 property parameter."""

    name: str

    values: list[Union[str, datetime.tzinfo]]
    """Values are typically strings, with a hack for TZID.

    The values may be overridden in the parse tree so that we can directly
    set the timezone information when parsing a date-time rather than
    combining with the calendar at runtime. That is, we update the tree
    with timezone infrmation replacing a string TZID with the zoneinfo.
    """


@dataclass
class ParsedProperty:
    """An rfc5545 property."""

    name: str
    value: str
    params: Optional[list[ParsedPropertyParameter]] = None

    def get_parameter(self, name: str) -> ParsedPropertyParameter | None:
        """Return a single ParsedPropertyParameter with the specified name."""
        if not self.params:
            return None
        for param in self.params:
            if param.name.lower() != name.lower():
                continue
            return param
        return None

    def get_parameter_value(self, name: str) -> str | None:
        """Return the property parameter value."""
        if not (param := self.get_parameter(name)):
            return None
        if len(param.values) > 1:
            raise ValueError(
                f"Expected only a single parameter string value, got {param.values}"
            )
        return param.values[0] if isinstance(param.values[0], str) else None

    def ics(self) -> str:
        """Encode a ParsedProperty into the serialized format."""
        result = [self.name.upper()]
        if self.params:
            result.append(";")
            result_params = []
            for parameter in self.params:
                result_param_values = []
                for value in parameter.values:
                    if not isinstance(value, str):
                        continue  # Shouldn't happen; only strings are set by parsing
                    # Property parameters with values contain a colon, simicolon,
                    # or a comma character must be placed in quoted text
                    if UNSAFE_CHAR_RE.search(value):
                        result_param_values.append(f'"{value}"')
                    else:
                        result_param_values.append(value)
                values = ",".join(result_param_values)
                result_params.append(f"{parameter.name.upper()}={values}")
            result.append(";".join(result_params))
        result.append(":")
        result.append(str(self.value))
        return "".join(result)


def parse_property_params(
    parse_result_dict: dict[str, str | list]
) -> list[ParsedPropertyParameter]:
    """Extract the property parameters from a pyparsing ParseResult object."""
    if PARSE_PARAMS not in parse_result_dict:
        return []
    params: list[ParsedPropertyParameter] = []
    for parsed_params in parse_result_dict[PARSE_PARAMS]:
        if not isinstance(parsed_params, dict) or PARSE_PARAMS not in parsed_params:
            continue
        for pair in parsed_params[PARSE_PARAMS]:
            params.append(
                ParsedPropertyParameter(
                    name=pair[PARSE_PARAM_NAME], values=pair[PARSE_PARAM_VALUE]
                )
            )
    return params



"""Tests for diagnostics."""

import yaml
import pytest

from ical.diagnostics import redact_ics


DATETIME_TIMEZONE = """BEGIN:VCALENDAR
PRODID:-//hacksw/handcal//NONSGML v1.0//EN
VERSION:***
BEGIN:VEVENT
DTSTAMP:19970610T172345Z
UID:***
DTSTART;TZID=America/New_York:19970714T133000
DTEND;TZID=America/New_York:19970714T140000
SUMMARY:***
END:VEVENT
END:VCALENDAR"""

DESCRIPTION_ALTREP = """BEGIN:VCALENDAR
PRODID:-//hacksw/handcal//NONSGML v1.0//EN
VERSION:***
BEGIN:VEVENT
DTSTAMP:19970610T172345Z
UID:***
DTSTART:19970714T170000Z
DTEND:19970715T040000Z
SUMMARY:***
DESCRIPTION:***
***
END:VEVENT
END:VCALENDAR"""


def test_empty() -> None:
    """Test redaction of an empty ics file."""
    assert list(redact_ics("")) == []
    assert list(redact_ics("\n")) == []


@pytest.mark.parametrize(
    ("filename", "output"),
    [
        ("tests/testdata/datetime_timezone.yaml", DATETIME_TIMEZONE),
        ("tests/testdata/description_altrep.yaml", DESCRIPTION_ALTREP),
    ],
    ids=("datetime_timezone", "description_altrep"),
)
def test_redact_date_timezone(filename: str, output: str) -> None:
    """Test redaction of an empty ics file."""

    with open(filename) as f:
        doc = yaml.load(f, Loader=yaml.CLoader)
    ics = doc["input"]
    assert "\n".join(list(redact_ics(ics))) == output



"""Library for parsing and encoding UTC-OFFSET values."""

from __future__ import annotations

import datetime
import re
from dataclasses import dataclass
from typing import Any

from ical.parsing.property import ParsedProperty

from .data_types import DATA_TYPE

UTC_OFFSET_REGEX = re.compile(r"^([-+]?)([0-9]{2})([0-9]{2})$")


@DATA_TYPE.register("UTC-OFFSET")
@dataclass
class UtcOffset:
    """Contains an offset from UTC to local time."""

    offset: datetime.timedelta

    @classmethod
    def __parse_property_value__(cls, prop: Any) -> UtcOffset:
        """Parse a UTC Offset."""
        if isinstance(prop, UtcOffset):
            return prop
        value = prop
        if isinstance(prop, ParsedProperty):
            value = prop.value
        if not (match := UTC_OFFSET_REGEX.fullmatch(value)):
            raise ValueError(f"Expected value to match UTC-OFFSET pattern: {value}")
        sign, hours, minutes = match.groups()
        result = datetime.timedelta(
            hours=int(hours or 0),
            minutes=int(minutes or 0),
        )
        if sign == "-":
            result = -result
        return UtcOffset(result)

    @classmethod
    def __encode_property_json__(cls, value: UtcOffset) -> str:
        """Serialize a time delta as a UTC-OFFSET ICS value."""
        duration = value.offset
        parts = []
        if duration < datetime.timedelta(days=0):
            parts.append("-")
            duration = -duration
        seconds = duration.seconds
        hours = int(seconds / 3600)
        seconds %= 3600
        parts.append(f"{hours:02}")
        minutes = int(seconds / 60)
        seconds %= 60
        parts.append(f"{minutes:02}")
        return "".join(parts)



"""A grouping of component properties that describe a to-do."""

from __future__ import annotations

import datetime
import enum
from typing import Any, Optional, Union

from pydantic import Field, root_validator

from .alarm import Alarm
from .component import ComponentModel, validate_until_dtstart
from .parsing.property import ParsedProperty
from .types import (
    CalAddress,
    Classification,
    Geo,
    Priority,
    Recur,
    RecurrenceId,
    RequestStatus,
    Uri,
)
from .util import dtstamp_factory, normalize_datetime, uid_factory


class TodoStatus(str, enum.Enum):
    """Status or confirmation of the to-do."""

    NEEDS_ACTION = "NEEDS-ACTION"
    COMPLETED = "COMPLETED"
    IN_PROCESS = "IN-PROCESS"
    CANCELLED = "CANCELLED"


class Todo(ComponentModel):
    """A calendar todo component."""

    dtstamp: Union[datetime.datetime, datetime.date] = Field(
        default_factory=dtstamp_factory
    )
    uid: str = Field(default_factory=uid_factory)

    attendees: list[CalAddress] = Field(alias="attendee", default_factory=list)
    categories: list[str] = Field(default_factory=list)
    classification: Optional[Classification] = Field(alias="class", default=None)
    comment: list[str] = Field(default_factory=list)
    completed: Optional[datetime.datetime] = None
    contacts: list[str] = Field(alias="contact", default_factory=list)
    created: Optional[datetime.datetime] = None
    description: Optional[str] = None

    # Has alias of 'start'
    dtstart: Union[datetime.datetime, datetime.date, None] = None

    due: Union[datetime.datetime, datetime.date, None] = None
    duration: Optional[datetime.timedelta] = None
    exdate: list[Union[datetime.datetime, datetime.date]] = Field(default_factory=list)
    geo: Optional[Geo] = None
    last_modified: Optional[datetime.datetime] = Field(
        alias="last-modified", default=None
    )
    location: str = ""
    organizer: Optional[CalAddress] = None
    percent: Optional[int] = None
    priority: Optional[Priority] = None
    recurrence_id: Optional[RecurrenceId] = Field(alias="recurrence-id")
    request_status: Optional[RequestStatus] = Field(
        alias="request-status",
        default_value=None,
    )
    rrule: Optional[Recur] = None
    sequence: Optional[int] = None
    status: Optional[TodoStatus] = None
    summary: Optional[str] = None
    url: Optional[Uri] = None

    alarms: list[Alarm] = Field(alias="valarm", default_factory=list)

    extras: list[ParsedProperty] = Field(default_factory=list)

    def __init__(self, **data: Any) -> None:
        """Initialize Todo."""
        if "start" in data:
            data["dtstart"] = data.pop("start")
        super().__init__(**data)

    @property
    def start(self) -> datetime.datetime | datetime.date | None:
        """Return the start time for the todo."""
        return self.dtstart

    @property
    def start_datetime(self) -> datetime.datetime | None:
        """Return the todos start as a datetime."""
        if not self.dtstart:
            return None
        return normalize_datetime(self.dtstart).astimezone(tz=datetime.timezone.utc)

    @root_validator
    def validate_one_due_or_duration(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that only one of duration or end date may be set."""
        if values.get("due") and values.get("duration"):
            raise ValueError("Only one of dtend or duration may be set." "")
        return values

    @root_validator
    def validate_duration_requires_start(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate that only one of duration or end date may be set."""
        if values.get("duration") and not values.get("dtstart"):
            raise ValueError("Duration requires that dtstart is specified")
        return values

    _validate_until_dtstart = root_validator(allow_reuse=True)(validate_until_dtstart)



"""Tests for timeline related calendar eents."""
from __future__ import annotations

import datetime
import re
import uuid
import zoneinfo
from typing import Generator
from unittest.mock import patch

import pytest
from freezegun import freeze_time

from ical.calendar import Calendar
from ical.calendar_stream import IcsCalendarStream
from ical.event import Event


@pytest.fixture(name="calendar")
def mock_calendar() -> Calendar:
    """Fixture calendar with all day events to use in tests."""
    cal = Calendar()
    cal.events.extend(
        [
            Event(
                summary="second",
                start=datetime.date(2000, 2, 1),
                end=datetime.date(2000, 2, 2),
            ),
            Event(
                summary="fourth",
                start=datetime.date(2000, 4, 1),
                end=datetime.date(2000, 4, 2),
            ),
            Event(
                summary="third",
                start=datetime.date(2000, 3, 1),
                end=datetime.date(2000, 3, 2),
            ),
            Event(
                summary="first",
                start=datetime.date(2000, 1, 1),
                end=datetime.date(2000, 1, 2),
            ),
        ]
    )
    return cal


@pytest.fixture(name="calendar_times")
def mock_calendar_times() -> Calendar:
    """Fixture calendar with datetime based events to use in tests."""
    cal = Calendar()
    cal.events.extend(
        [
            Event(
                summary="first",
                start=datetime.datetime(2000, 1, 1, 11, 0),
                end=datetime.datetime(2000, 1, 1, 11, 30),
            ),
            Event(
                summary="second",
                start=datetime.datetime(2000, 1, 1, 12, 0),
                end=datetime.datetime(2000, 1, 1, 13, 0),
            ),
            Event(
                summary="third",
                start=datetime.datetime(2000, 1, 2, 12, 0),
                end=datetime.datetime(2000, 1, 2, 13, 0),
            ),
        ]
    )
    return cal


@pytest.fixture(name="calendar_mixed")
def mock_calendar_mixed() -> Calendar:
    """Fixture with a mix of all day and datetime based events."""
    cal = Calendar()
    cal.events.extend(
        [
            Event(
                summary="All Day",
                start=datetime.date(2000, 2, 1),
                end=datetime.date(2000, 2, 2),
            ),
            Event(
                summary="Event @ 7 UTC",
                start=datetime.datetime(
                    2000, 2, 1, 7, 00, 0, tzinfo=datetime.timezone.utc
                ),
                end=datetime.datetime(
                    2000, 2, 2, 7, 00, 0, tzinfo=datetime.timezone.utc
                ),
            ),
            Event(
                summary="Event @ 5UTC",
                start=datetime.datetime(
                    2000, 2, 1, 5, 00, 0, tzinfo=datetime.timezone.utc
                ),
                end=datetime.datetime(
                    2000, 2, 2, 5, 00, 0, tzinfo=datetime.timezone.utc
                ),
            ),
        ]
    )
    return cal


def test_iteration(calendar: Calendar) -> None:
    """Test chronological iteration of a timeline."""
    assert [e.summary for e in calendar.timeline] == [
        "first",
        "second",
        "third",
        "fourth",
    ]


@pytest.mark.parametrize(
    "when,expected_events",
    [
        (datetime.date(2000, 1, 1), ["first"]),
        (datetime.date(2000, 2, 1), ["second"]),
        (datetime.date(2000, 3, 1), ["third"]),
    ],
)
def test_on_date(
    calendar: Calendar, when: datetime.date, expected_events: list[str]
) -> None:
    """Test returning events on a particular day."""
    assert [e.summary for e in calendar.timeline.on_date(when)] == expected_events


def test_start_after(calendar: Calendar) -> None:
    """Test chronological iteration starting at a specific time."""
    assert [
        e.summary for e in calendar.timeline.start_after(datetime.date(2000, 1, 1))
    ] == ["second", "third", "fourth"]
    assert [
        e.summary
        for e in calendar.timeline.start_after(datetime.datetime(2000, 1, 1, 6, 0, 0))
    ] == ["second", "third", "fourth"]
    assert [
        e.summary
        for e in calendar.timeline.start_after(datetime.datetime(2000, 1, 15, 0, 0, 0))
    ] == ["second", "third", "fourth"]


def test_start_after_times(calendar_times: Calendar) -> None:
    """Test chronological iteration starting at a specific time."""
    assert [
        e.summary
        for e in calendar_times.timeline.start_after(datetime.date(2000, 1, 1))
    ] == ["first", "second", "third"]
    assert [
        e.summary
        for e in calendar_times.timeline.start_after(
            datetime.datetime(2000, 1, 1, 6, 0, 0)
        )
    ] == ["first", "second", "third"]


def test_active_after(calendar: Calendar) -> None:
    """Test chronological iteration starting at a specific time."""
    assert [
        e.summary
        for e in calendar.timeline.start_after(datetime.datetime(2000, 1, 1, 12, 0, 0))
    ] == ["second", "third", "fourth"]
    assert [
        e.summary for e in calendar.timeline.start_after(datetime.date(2000, 1, 1))
    ] == ["second", "third", "fourth"]


def test_active_after_times(calendar_times: Calendar) -> None:
    """Test chronological iteration starting at a specific time."""
    assert [
        e.summary
        for e in calendar_times.timeline.start_after(
            datetime.datetime(2000, 1, 1, 12, 0, 0)
        )
    ] == ["third"]
    assert [
        e.summary
        for e in calendar_times.timeline.start_after(datetime.date(2000, 1, 1))
    ] == ["first", "second", "third"]


@pytest.mark.parametrize(
    "at_datetime,expected_events",
    [
        (datetime.datetime(2000, 1, 1, 11, 15), ["first"]),
        (datetime.datetime(2000, 1, 1, 11, 59), []),
        (datetime.datetime(2000, 1, 1, 12, 0), ["second"]),
        (datetime.datetime(2000, 1, 1, 12, 30), ["second"]),
        (datetime.datetime(2000, 1, 1, 12, 59), ["second"]),
        (datetime.datetime(2000, 1, 1, 13, 0), []),
    ],
)
def test_at_instant(
    calendar_times: Calendar, at_datetime: datetime.datetime, expected_events: list[str]
) -> None:
    """Test returning events at a specific time."""
    assert [
        e.summary for e in calendar_times.timeline.at_instant(at_datetime)
    ] == expected_events


@freeze_time("2000-01-01 12:30:00")
def test_now(calendar_times: Calendar) -> None:
    """Test events happening at the current time."""
    assert [e.summary for e in calendar_times.timeline.now()] == ["second"]


@freeze_time("2000-01-01 13:00:00")
def test_now_no_match(calendar_times: Calendar) -> None:
    """Test no events happening at the current time."""
    assert [e.summary for e in calendar_times.timeline.now()] == []


@freeze_time("2000-01-01 12:30:00")
def test_today(calendar_times: Calendar) -> None:
    """Test events active today."""
    assert [e.summary for e in calendar_times.timeline.today()] == ["first", "second"]


@pytest.mark.parametrize(
    "start,end,expected_events",
    [
        (
            datetime.datetime(2000, 1, 1, 10, 00),
            datetime.datetime(2000, 1, 2, 14, 00),
            ["first", "second", "third"],
        ),
        (
            datetime.datetime(2000, 1, 1, 10, 00),
            datetime.datetime(2000, 1, 1, 14, 00),
            ["first", "second"],
        ),
        (
            datetime.datetime(2000, 1, 1, 12, 00),
            datetime.datetime(2000, 1, 2, 14, 00),
            ["second", "third"],
        ),
        (
            datetime.datetime(2000, 1, 1, 12, 00),
            datetime.datetime(2000, 1, 1, 14, 00),
            ["second"],
        ),
    ],
)
def test_included(
    calendar_times: Calendar,
    start: datetime.datetime,
    end: datetime.datetime,
    expected_events: list[str],
) -> None:
    """Test calendar timeline inclusions."""
    assert [
        e.summary for e in calendar_times.timeline.included(start, end)
    ] == expected_events


def test_multiple_calendars(calendar: Calendar, calendar_times: Calendar) -> None:
    """Test multiple calendars have independent event lists."""
    assert len(calendar.events) == 4
    assert len(calendar_times.events) == 3
    assert len(Calendar().events) == 0


def test_multiple_iteration(calendar: Calendar) -> None:
    """Test iterating over a timeline multiple times."""
    line = calendar.timeline
    assert [e.summary for e in line] == [
        "first",
        "second",
        "third",
        "fourth",
    ]
    assert [e.summary for e in line] == [
        "first",
        "second",
        "third",
        "fourth",
    ]


TEST_ICS = """BEGIN:VCALENDAR
PRODID:-//hacksw/handcal//NONSGML v1.0//EN
VERSION:2.0
BEGIN:VEVENT
DTSTAMP:19970610T172345Z
UID:19970610T172345Z-AF23B2@example.com
DTSTART:19970714T170000Z
DTEND:19970715T040000Z
SUMMARY:Bastille Day Party
END:VEVENT
END:VCALENDAR"""


def test_calendar_serialization() -> None:
    """Test single calendar serialization."""
    calendar = IcsCalendarStream.calendar_from_ics(TEST_ICS)
    assert len(calendar.events) == 1
    assert calendar.events[0].summary == "Bastille Day Party"
    output_ics = IcsCalendarStream.calendar_to_ics(calendar)
    assert output_ics == TEST_ICS


def test_empty_calendar() -> None:
    """Test reading an empty calendar file."""
    calendar = IcsCalendarStream.calendar_from_ics("")
    assert len(calendar.events) == 0


@pytest.fixture(name="_uid")
def mock_uid() -> Generator[str, None, None]:
    """Patch out uuid creation with a fixed value."""
    value = str(uuid.uuid3(uuid.NAMESPACE_DNS, "fixed-name"))
    with patch(
        "ical.event.uid_factory",
        return_value=value,
    ):
        yield value


@freeze_time("2000-01-01 12:30:00")
def test_create_and_serialize_calendar(
    _uid: str,
) -> None:
    """Test creating a calendar manually then serializing."""
    cal = Calendar()
    cal.events.append(
        Event(
            summary="Event",
            start=datetime.date(2000, 2, 1),
            end=datetime.date(2000, 2, 2),
        )
    )
    ics = IcsCalendarStream.calendar_to_ics(cal)
    assert re.split("\r?\n", ics) == [
        "BEGIN:VCALENDAR",
        "BEGIN:VEVENT",
        "DTSTAMP:20000101T123000",
        "UID:68e9e07c-7557-36e2-91c1-8febe7527841",
        "DTSTART:20000201",
        "DTEND:20000202",
        "SUMMARY:Event",
        "END:VEVENT",
        "END:VCALENDAR",
    ]


def test_mixed_iteration_order(calendar_mixed: Calendar) -> None:
    """Test iteration order of all day events based on the attendee timezone."""

    # UTC order
    assert [e.summary for e in calendar_mixed.timeline_tz(datetime.timezone.utc)] == [
        "All Day",
        "Event @ 5UTC",
        "Event @ 7 UTC",
    ]

    # Attendee is in -6
    assert [
        e.summary
        for e in calendar_mixed.timeline_tz(zoneinfo.ZoneInfo("America/Regina"))
    ] == ["Event @ 5UTC", "All Day", "Event @ 7 UTC"]

    # Attendee is in -8
    assert [
        e.summary
        for e in calendar_mixed.timeline_tz(zoneinfo.ZoneInfo("America/Los_Angeles"))
    ] == ["Event @ 5UTC", "Event @ 7 UTC", "All Day"]


@pytest.mark.parametrize(
    "tzname,dt_before,dt_after",
    [
        (
            "America/Los_Angeles",  # UTC-8 in Feb
            datetime.datetime(2000, 2, 1, 7, 59, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(2000, 2, 1, 8, 0, 0, tzinfo=datetime.timezone.utc),
        ),
        (
            "America/Regina",  # UTC-6 all year round
            datetime.datetime(2000, 2, 1, 5, 59, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(2000, 2, 1, 6, 0, 0, tzinfo=datetime.timezone.utc),
        ),
        (
            "CET",  # UTC-1 in Feb
            datetime.datetime(2000, 1, 31, 22, 59, 59, tzinfo=datetime.timezone.utc),
            datetime.datetime(2000, 1, 31, 23, 0, 0, tzinfo=datetime.timezone.utc),
        ),
    ],
)
def test_all_day_with_local_timezone(
    tzname: str, dt_before: datetime.datetime, dt_after: datetime.datetime
) -> None:
    """Test iteration of all day events using local timezone override."""
    cal = Calendar()
    cal.events.extend(
        [
            Event(
                summary="event",
                start=datetime.date(2000, 2, 1),
                end=datetime.date(2000, 2, 2),
            ),
        ]
    )

    local_tz = zoneinfo.ZoneInfo(tzname)

    def start_after(dtstart: datetime.datetime) -> list[str]:
        nonlocal cal, local_tz
        return [e.summary for e in cal.timeline_tz(local_tz).start_after(dtstart)]

    local_before = dt_before.astimezone(local_tz)
    assert start_after(local_before) == ["event"]

    local_after = dt_after.astimezone(local_tz)
    assert not start_after(local_after)



"""Tests for Event component."""

from __future__ import annotations

import zoneinfo
from datetime import date, datetime, timedelta, timezone
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from ical.event import Event

SUMMARY = "test summary"
LOS_ANGELES = zoneinfo.ZoneInfo("America/Los_Angeles")


@pytest.mark.parametrize(
    "begin,end,duration",
    [
        (
            datetime.fromisoformat("2022-09-16 12:00"),
            datetime.fromisoformat("2022-09-16 12:30"),
            timedelta(minutes=30),
        ),
        (
            date.fromisoformat("2022-09-16"),
            date.fromisoformat("2022-09-17"),
            timedelta(hours=24),
        ),
        (
            datetime.fromisoformat("2022-09-16 06:00"),
            datetime.fromisoformat("2022-09-17 08:30"),
            timedelta(days=1, hours=2, minutes=30),
        ),
    ],
)
def test_start_end_duration(
    begin: datetime, end: datetime, duration: timedelta
) -> None:
    """Test event duration calculation."""
    event = Event(summary=SUMMARY, start=begin, end=end)
    assert event.computed_duration == duration
    assert not event.duration


@pytest.mark.parametrize(
    "event1_start,event1_end,event2_start,event2_end",
    [
        (date(2022, 9, 6), date(2022, 9, 7), date(2022, 9, 8), date(2022, 9, 10)),
        (
            datetime(2022, 9, 6, 6, 0, 0),
            datetime(2022, 9, 6, 7, 0, 0),
            datetime(2022, 9, 6, 8, 0, 0),
            datetime(2022, 9, 6, 8, 30, 0),
        ),
        (
            datetime(2022, 9, 6, 6, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 6, 7, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 6, 8, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 6, 8, 30, 0, tzinfo=timezone.utc),
        ),
        (
            datetime(2022, 9, 6, 6, 0, 0, tzinfo=LOS_ANGELES),
            datetime(2022, 9, 6, 7, 0, 0, tzinfo=LOS_ANGELES),
            datetime(2022, 9, 7, 8, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 7, 8, 30, 0, tzinfo=timezone.utc),
        ),
        (
            datetime(2022, 9, 6, 6, 0, 0, tzinfo=LOS_ANGELES),
            datetime(2022, 9, 6, 7, 0, 0, tzinfo=LOS_ANGELES),
            datetime(2022, 9, 8, 8, 0, 0),
            datetime(2022, 9, 8, 8, 30, 0),
        ),
        (
            datetime(2022, 9, 6, 6, 0, 0, tzinfo=LOS_ANGELES),
            datetime(2022, 9, 6, 7, 0, 0, tzinfo=LOS_ANGELES),
            date(2022, 9, 8),
            date(2022, 9, 9),
        ),
        (
            date(2022, 9, 6),
            date(2022, 9, 7),
            datetime(2022, 9, 6, 8, 0, 1, tzinfo=timezone.utc),
            datetime(2022, 9, 6, 8, 30, 0, tzinfo=timezone.utc),
        ),
    ],
)
def test_comparisons(
    event1_start: datetime | date,
    event1_end: datetime | date,
    event2_start: datetime | date,
    event2_end: datetime | date,
) -> None:
    """Test event comparison methods."""
    event1 = Event(summary=SUMMARY, start=event1_start, end=event1_end)
    event2 = Event(summary=SUMMARY, start=event2_start, end=event2_end)
    assert event1 < event2
    assert event1 <= event2
    assert event2 >= event1
    assert event2 > event1

    assert event1 <= event2
    assert event2 >= event1
    assert event2 > event1


def test_invalid_comparisons() -> None:
    """Test event comparisons that are not valid."""
    event1 = Event(summary=SUMMARY, start=date(2022, 9, 6), end=date(2022, 9, 7))

    with pytest.raises(TypeError):
        assert event1 < "example"

    with pytest.raises(TypeError):
        assert event1 <= "example"

    with pytest.raises(TypeError):
        assert event1 > "example"

    with pytest.raises(TypeError):
        assert event1 >= "example"


def test_within_and_includes() -> None:
    """Test more complex comparison methods."""
    event1 = Event(summary=SUMMARY, start=date(2022, 9, 6), end=date(2022, 9, 10))
    event2 = Event(summary=SUMMARY, start=date(2022, 9, 7), end=date(2022, 9, 8))
    event3 = Event(summary=SUMMARY, start=date(2022, 9, 9), end=date(2022, 9, 11))

    assert not event1.starts_within(event2)
    assert not event1.starts_within(event3)
    assert event2.starts_within(event1)
    assert not event2.starts_within(event3)
    assert event3.starts_within(event1)
    assert not event3.starts_within(event2)

    assert not event1.ends_within(event2)
    assert event1.ends_within(event3)
    assert event2.ends_within(event1)
    assert not event2.ends_within(event3)
    assert not event3.ends_within(event1)
    assert not event3.ends_within(event2)
    assert event2 > event1

    assert event1.includes(event2)
    assert not event1.includes(event3)
    assert not event2.includes(event1)
    assert not event2.includes(event3)
    assert not event3.includes(event1)
    assert not event3.includes(event2)

    assert not event1.is_included_in(event2)
    assert not event1.is_included_in(event3)
    assert event2.is_included_in(event1)
    assert not event2.is_included_in(event3)
    assert not event3.is_included_in(event1)
    assert not event3.is_included_in(event2)


def test_start_end_same_type() -> None:
    """Verify that the start and end value are the same type."""

    with pytest.raises(ValidationError):
        Event(
            summary=SUMMARY, start=date(2022, 9, 9), end=datetime(2022, 9, 9, 11, 0, 0)
        )

    with pytest.raises(ValidationError):
        Event(
            summary=SUMMARY, start=datetime(2022, 9, 9, 10, 0, 0), end=date(2022, 9, 9)
        )


def test_start_end_local_time() -> None:
    """Verify that the start and end value are the same type."""

    # Valid
    Event(
        summary=SUMMARY,
        start=datetime(2022, 9, 9, 10, 0, 0),
        end=datetime(2022, 9, 9, 11, 0, 0),
    )
    Event(
        summary=SUMMARY,
        start=datetime(2022, 9, 9, 10, 0, 0, tzinfo=timezone.utc),
        end=datetime(2022, 9, 9, 11, 0, 0, tzinfo=timezone.utc),
    )

    with pytest.raises(ValidationError):
        Event(
            summary=SUMMARY,
            start=datetime(2022, 9, 9, 10, 0, 0, tzinfo=timezone.utc),
            end=datetime(2022, 9, 9, 11, 0, 0),
        )

    with pytest.raises(ValidationError):
        Event(
            summary=SUMMARY,
            start=datetime(2022, 9, 9, 10, 0, 0),
            end=datetime(2022, 9, 9, 11, 0, 0, tzinfo=timezone.utc),
        )


def test_start_and_duration() -> None:
    """Verify event created with a duration instead of explicit end time."""

    event = Event(summary=SUMMARY, start=date(2022, 9, 9), duration=timedelta(days=3))
    assert event.start == date(2022, 9, 9)
    assert event.end == date(2022, 9, 12)

    with pytest.raises(ValidationError):
        Event(summary=SUMMARY, start=date(2022, 9, 9), duration=timedelta(days=-3))

    with pytest.raises(ValidationError):
        Event(summary=SUMMARY, start=date(2022, 9, 9), duration=timedelta(seconds=60))

    event = Event(
        summary=SUMMARY,
        start=datetime(2022, 9, 9, 10, 0, 0),
        duration=timedelta(seconds=60),
    )
    assert event.start == datetime(2022, 9, 9, 10, 0, 0)
    assert event.end == datetime(2022, 9, 9, 10, 1, 0)
    assert event.duration == timedelta(seconds=60)
    assert event.computed_duration == timedelta(seconds=60)


@pytest.mark.parametrize(
    "range1,range2,expected",
    [
        (
            (date(2022, 8, 1), date(2022, 8, 2)),
            (date(2022, 8, 1), date(2022, 8, 2)),
            True,
        ),
        (
            (date(2022, 8, 1), date(2022, 8, 4)),
            (date(2022, 8, 2), date(2022, 8, 3)),
            True,
        ),
        (
            (date(2022, 8, 1), date(2022, 8, 3)),
            (date(2022, 8, 2), date(2022, 8, 4)),
            True,
        ),
        (
            (date(2022, 8, 2), date(2022, 8, 3)),
            (date(2022, 8, 1), date(2022, 8, 4)),
            True,
        ),
        (
            (date(2022, 8, 3), date(2022, 8, 4)),
            (date(2022, 8, 1), date(2022, 8, 4)),
            True,
        ),
        (
            (date(2022, 8, 2), date(2022, 8, 4)),
            (date(2022, 8, 1), date(2022, 8, 4)),
            True,
        ),
        (
            (date(2022, 8, 1), date(2022, 8, 2)),
            (date(2022, 8, 3), date(2022, 8, 4)),
            False,
        ),
        (
            (date(2022, 8, 3), date(2022, 8, 4)),
            (date(2022, 8, 1), date(2022, 8, 2)),
            False,
        ),
        (
            (date(2022, 8, 1), date(2022, 8, 2)),
            (date(2022, 8, 2), date(2022, 8, 3)),
            False,
        ),
        (
            (date(2022, 8, 2), date(2022, 8, 3)),
            (date(2022, 8, 1), date(2022, 8, 2)),
            False,
        ),
    ],
)
def test_date_intersects(
    range1: tuple[date, date],
    range2: tuple[date, date],
    expected: bool,
) -> None:
    """Test event intersection."""
    event1 = Event(summary=SUMMARY, start=range1[0], end=range1[1])
    event2 = Event(summary=SUMMARY, start=range2[0], end=range2[1])
    assert event1.intersects(event2) == expected


@pytest.mark.parametrize(
    "start_str,end_str,start,end",
    [
        (
            "2022-09-16 12:00",
            "2022-09-16 12:30",
            datetime(2022, 9, 16, 12, 0, 0),
            datetime(2022, 9, 16, 12, 30, 0),
        ),
        (
            "2022-09-16",
            "2022-09-17",
            date(2022, 9, 16),
            date(2022, 9, 17),
        ),
        (
            "2022-09-16 06:00",
            "2022-09-17 08:30",
            datetime(2022, 9, 16, 6, 0, 0),
            datetime(2022, 9, 17, 8, 30, 0),
        ),
        (
            "2022-09-16T06:00",
            "2022-09-17T08:30",
            datetime(2022, 9, 16, 6, 0, 0),
            datetime(2022, 9, 17, 8, 30, 0),
        ),
        (
            "2022-09-16T06:00Z",
            "2022-09-17T08:30Z",
            datetime(2022, 9, 16, 6, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 17, 8, 30, 0, tzinfo=timezone.utc),
        ),
        (
            "2022-09-16T06:00+00:00",
            "2022-09-17T08:30+00:00",
            datetime(2022, 9, 16, 6, 0, 0, tzinfo=timezone.utc),
            datetime(2022, 9, 17, 8, 30, 0, tzinfo=timezone.utc),
        ),
        (
            "2022-09-16T06:00-07:00",
            "2022-09-17T08:30-07:00",
            datetime(2022, 9, 16, 6, 0, 0, tzinfo=timezone(offset=timedelta(hours=-7))),
            datetime(
                2022, 9, 17, 8, 30, 0, tzinfo=timezone(offset=timedelta(hours=-7))
            ),
        ),
    ],
)
def test_parse_event_timezones(
    start_str: str, end_str: str, start: datetime | date, end: datetime | date
) -> None:
    """Test parsing date/times from strings."""
    event = Event.parse_obj(
        {
            "summary": SUMMARY,
            "start": start_str,
            "end": end_str,
        }
    )
    assert event.start == start
    assert event.end == end


def test_all_day_timezones_default() -> None:
    """Test behavior of all day events interacting with timezones."""
    with patch(
        "ical.util.local_timezone", return_value=zoneinfo.ZoneInfo("America/Regina")
    ):
        event = Event(summary=SUMMARY, start=date(2022, 8, 1), end=date(2022, 8, 2))
        assert event.start_datetime == datetime(
            2022, 8, 1, 6, 0, 0, tzinfo=timezone.utc
        )
        assert event.end_datetime == datetime(2022, 8, 2, 6, 0, 0, tzinfo=timezone.utc)


@pytest.mark.parametrize(
    "dtstart,dtend",
    [
        (
            datetime(2022, 8, 1, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Regina")),
            datetime(2022, 8, 2, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Regina")),
        ),
        (
            datetime(
                2022, 8, 1, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
            datetime(
                2022, 8, 2, 0, 0, 0, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")
            ),
        ),
    ],
)
def test_all_day_timespan_timezone_explicit(dtstart: datetime, dtend: datetime) -> None:
    """Test behavior of all day events interacting with timezones specified explicitly."""
    event = Event(summary=SUMMARY, start=date(2022, 8, 1), end=date(2022, 8, 2))
    assert dtstart.tzinfo
    timespan = event.timespan_of(dtstart.tzinfo)
    assert timespan.start == dtstart
    assert timespan.end == dtend


def test_validate_assignment() -> None:
    """Test date type validations."""

    event = Event(summary=SUMMARY, start=date(2022, 9, 6), end=date(2022, 9, 7))

    # Validation on assignment ensures the start/end types can't be mismatched
    with pytest.raises(ValidationError):
        event.dtstart = datetime(2022, 9, 6, 6, 0, 0)
    with pytest.raises(ValidationError):
        event.dtend = datetime(2022, 9, 10, 6, 0, 0)

    # But updates that are valid are OK
    event.dtstart = date(2022, 9, 5)
    event.dtend = date(2022, 9, 10)



"""Tests for Todo component."""

from __future__ import annotations

import datetime
import zoneinfo
from unittest.mock import patch

import pytest
from pydantic import ValidationError

from ical.todo import Todo


def test_empty() -> None:
    """Test that in practice a Todo requires no fields."""
    todo = Todo()
    assert not todo.summary


def test_todo() -> None:
    """Test a valid Todo object."""
    todo = Todo(summary="Example", due=datetime.date(2022, 8, 7))
    assert todo.summary == "Example"
    assert todo.due == datetime.date(2022, 8, 7)


def test_duration() -> None:
    """Test relationship between the due and duration fields."""

    todo = Todo(start=datetime.date(2022, 8, 7), duration=datetime.timedelta(days=1))
    assert todo.start
    assert todo.duration

    # Both due and Duration can't be set
    with pytest.raises(ValidationError):
        Todo(
            start=datetime.date(2022, 8, 7),
            duration=datetime.timedelta(days=1),
            due=datetime.date(2022, 8, 8),
        )

    # Duration requires start date
    with pytest.raises(ValidationError):
        Todo(duration=datetime.timedelta(days=1))

    todo = Todo(start=datetime.date(2022, 8, 7), due=datetime.date(2022, 8, 8))
    assert todo.start
    assert todo.due
    assert todo.start_datetime

    with patch(
        "ical.util.local_timezone", return_value=zoneinfo.ZoneInfo("America/Regina")
    ):
        assert todo.start_datetime.isoformat() == "2022-08-07T06:00:00+00:00"



"""Library for parsing and encoding FLOAT values."""

from ical.parsing.property import ParsedProperty

from .data_types import DATA_TYPE


@DATA_TYPE.register("FLOAT")
class FloatEncoder:
    """Encode a float ICS value."""

    @classmethod
    def __property_type__(cls) -> type:
        return float

    @classmethod
    def __parse_property_value__(cls, prop: ParsedProperty) -> float:
        """Parse a rfc5545 property into a text value."""
        return float(prop.value)



"""Library for parsing TEXT values."""

from ical.parsing.property import ParsedProperty

from .data_types import DATA_TYPE

UNESCAPE_CHAR = {"\\\\": "\\", "\\;": ";", "\\,": ",", "\\N": "\n", "\\n": "\n"}
ESCAPE_CHAR = {v: k for k, v in UNESCAPE_CHAR.items()}


@DATA_TYPE.register("TEXT")
class TextEncoder:
    """Encode an rfc5545 TEXT value."""

    @classmethod
    def __property_type__(cls) -> type:
        return str

    @classmethod
    def __parse_property_value__(cls, prop: ParsedProperty) -> str:
        """Parse a rfc5545 into a text value."""
        for key, vin in UNESCAPE_CHAR.items():
            if key not in prop.value:
                continue
            prop.value = prop.value.replace(key, vin)
        return prop.value

    @classmethod
    def __encode_property_value__(cls, value: str) -> str:
        """Serialize text as an ICS value."""
        for key, vin in ESCAPE_CHAR.items():
            if key not in value:
                continue
            value = value.replace(key, vin)
        return value



"""Implementation of the REQUEST-STATUS property."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from .data_types import DATA_TYPE
from .text import TextEncoder


@dataclass
@DATA_TYPE.register()
class RequestStatus:
    """Status code returned for a scheduling request."""

    statcode: float
    statdesc: str
    exdata: Optional[str] = None

    @classmethod
    def __parse_property_value__(cls, value: Any) -> RequestStatus:
        """Parse a rfc5545 request status value."""
        parts = TextEncoder.__parse_property_value__(value).split(";")
        if len(parts) < 2 or len(parts) > 3:
            raise ValueError(f"Value was not valid Request Status: {value}")
        exdata: str | None = None
        if len(parts) == 3:
            exdata = parts[2]
        return RequestStatus(
            statcode=float(parts[0]),
            statdesc=parts[1],
            exdata=exdata,
        )

    @classmethod
    def __encode_property_json__(cls, value: RequestStatus) -> str:
        """Encoded RequestStatus as an ICS property."""
        result = f"{value.statcode};{value.statdesc}"
        if value.exdata:
            result += f";{value.exdata}"
        return result



"""Library for parsing and encoding DURATION values."""

import datetime
import re

from ical.parsing.property import ParsedProperty

from .data_types import DATA_TYPE

DATE_PART = r"(\d+)D"
TIME_PART = r"T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+)S)?"
DATETIME_PART = f"(?:{DATE_PART})?(?:{TIME_PART})?"
WEEKS_PART = r"(\d+)W"
DURATION_REGEX = re.compile(f"([-+]?)P(?:{WEEKS_PART}|{DATETIME_PART})$")


@DATA_TYPE.register("DURATION")
class DurationEncoder:
    """Class that can encode DURATION values."""

    @classmethod
    def __property_type__(cls) -> type:
        return datetime.timedelta

    @classmethod
    def __parse_property_value__(cls, prop: ParsedProperty) -> datetime.timedelta:
        """Parse a rfc5545 into a datetime.date."""
        if not isinstance(prop, ParsedProperty):
            raise ValueError(f"Expected ParsedProperty but was {prop}")
        if not (match := DURATION_REGEX.fullmatch(prop.value)):
            raise ValueError(f"Expected value to match DURATION pattern: {prop.value}")
        sign, weeks, days, hours, minutes, seconds = match.groups()
        result: datetime.timedelta
        if weeks:
            result = datetime.timedelta(weeks=int(weeks))
        else:
            result = datetime.timedelta(
                days=int(days or 0),
                hours=int(hours or 0),
                minutes=int(minutes or 0),
                seconds=int(seconds or 0),
            )
        if sign == "-":
            result = -result
        return result

    @classmethod
    def __encode_property_json__(cls, duration: datetime.timedelta) -> str:
        """Serialize a time delta as a DURATION ICS value."""
        parts = []
        if duration < datetime.timedelta(days=0):
            parts.append("-")
            duration = -duration
        parts.append("P")
        days = duration.days
        weeks = int(days / 7)
        days %= 7
        if weeks > 0:
            parts.append(f"{weeks}W")
        if days > 0:
            parts.append(f"{days}D")
        if duration.seconds != 0:
            parts.append("T")
            seconds = duration.seconds
            hours = int(seconds / 3600)
            seconds %= 3600
            if hours != 0:
                parts.append(f"{hours}H")
            minutes = int(seconds / 60)
            seconds %= 60
            if minutes != 0:
                parts.append(f"{minutes}M")
            if seconds != 0:
                parts.append(f"{seconds}S")
        return "".join(parts)



"""The Calendar component."""

from __future__ import annotations

import datetime
from importlib import metadata
import itertools
import logging
from typing import Optional, Any
import zoneinfo

from pydantic import Field, root_validator

from .component import ComponentModel
from .event import Event
from .freebusy import FreeBusy
from .journal import Journal
from .types.date_time import TZID
from .parsing.property import ParsedProperty
from .timeline import Timeline, calendar_timeline
from .timezone import Timezone, TimezoneModel, IcsTimezoneInfo
from .todo import Todo
from .util import local_timezone


_LOGGER = logging.getLogger(__name__)


_VERSION = metadata.version("ical")
_PRODID = metadata.metadata("ical")["prodid"]

# Components that may contain TZID objects
_TZID_COMPONENTS = ["vevent", "vtodo", "vjournal", "vfreebusy"]


class Calendar(ComponentModel):
    """A sequence of calendar properties and calendar components."""

    calscale: Optional[str] = None
    method: Optional[str] = None
    prodid: str = Field(default=_PRODID)
    version: str = Field(default=_VERSION)

    #
    # Calendar components
    #

    events: list[Event] = Field(alias="vevent", default_factory=list)
    """Events associated with this calendar."""

    todos: list[Todo] = Field(alias="vtodo", default_factory=list)
    """Todos associated with this calendar."""

    journal: list[Journal] = Field(alias="vjournal", default_factory=list)
    """Journals associated with this calendar."""

    freebusy: list[FreeBusy] = Field(alias="vfreebusy", default_factory=list)
    """Free/busy objects associated with this calendar."""

    timezones: list[Timezone] = Field(alias="vtimezone", default_factory=list)
    """Timezones associated with this calendar."""

    # Unknown or unsupported properties
    extras: list[ParsedProperty] = Field(default_factory=list)

    @property
    def timeline(self) -> Timeline:
        """Return a timeline view of events on the calendar.

        All day events are returned as if the attendee is viewing from UTC time.
        """
        return self.timeline_tz()

    def timeline_tz(self, tzinfo: datetime.tzinfo | None = None) -> Timeline:
        """Return a timeline view of events on the calendar.

        All events are returned as if the attendee is viewing from the
        specified timezone. For example, this affects the order that All Day
        events are returned.
        """
        return calendar_timeline(self.events, tzinfo=tzinfo or local_timezone())

    @root_validator(pre=True)
    def _propagate_timezones(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Propagate timezone information down to date-time objects.

        This results in parsing the timezones twice: Once here, then once again
        in the calendar itself. This does imply that if a vtimezone object is
        changed live, the DATE-TIME objects are not updated.

        We first update the timezone objects using another pydantic model just
        for parsing and propagaint here (TimezoneModel). We then walk through
        all DATE-TIME objects referenced by components and lookup any TZID
        property parameters, converting them to a datetime.tzinfo object. The
        DATE-TIME parser will use this instead of the TZID string. We prefer
        to use any python timezones when present.
        """

        # First parse the timezones out of the calendar, ignoring everything else
        timezone_model = TimezoneModel.parse_obj(values)
        system_tzids = zoneinfo.available_timezones()
        tzinfos: dict[str, datetime.tzinfo] = {
            timezone.tz_id: IcsTimezoneInfo.from_timezone(timezone)
            for timezone in timezone_model.timezones
            if timezone.tz_id not in system_tzids
        }
        if not timezone_model.timezones:
            return values
        # Replace any TZID objects with a reference to tzinfo from this calendar. The
        # DATE-TIME parser will use that if present.
        _LOGGER.debug("Replacing timezone (num %d) references in events", len(tzinfos))
        components = itertools.chain.from_iterable(
            [values.get(component, []) for component in _TZID_COMPONENTS]
        )
        for event in components:
            for field_values in event.values():
                for value in field_values or []:
                    if not isinstance(value, ParsedProperty):
                        continue
                    if (
                        not (tzid_param := value.get_parameter(TZID))
                        or not tzid_param.values
                    ):
                        continue
                    if isinstance(tzid_param.values[0], str) and (
                        tzinfo := tzinfos.get(tzid_param.values[0])
                    ):
                        tzid_param.values = [tzinfo]
        return values



"""Character sets used inused in rfc5545.

This file defines the character sets used by pyparsing to tokenize
input files, intended to be used by the contentlines parsing code.
"""

from __future__ import annotations

import logging
import re
from typing import cast

import emoji
from pyparsing import unicode_set
from pyparsing.unicode import UnicodeRangeList

_LOGGER = logging.getLogger(__name__)


WSP = [" ", "\t"]


class CharRange(unicode_set):
    """A base class that returns all characters in a range."""

    @classmethod
    def all(cls) -> list[str]:
        """Return all characters from the range."""
        return cast(list[str], cls._chars_for_ranges)


class Control(CharRange):
    """All the controls excet HTAB."""

    _ranges: UnicodeRangeList = [
        (0x00, 0x08),
        (0x0A, 0x1F),
        (0x7F,),
    ]


class NonUsAscii(CharRange):
    """Range of characters in the NON-US-ASCII spec of rfc5545."""

    _ranges: UnicodeRangeList = [
        # UTF8-2
        (0xC2, 0xDF),
        (0x80, 0xBF),  # utf8-tail
        # UTF8-3
        (0xE0,),
        (0xA0, 0xBF),
        (0xE1, 0xEC),
        (0xED,),
        (0x80, 0x9F),
        (0xEE, 0xEF),
        # UTF8-4
        (0xF0,),
        (0x90, 0xBF),
        (0xF1, 0xF3),
        (0xF4,),
        (0x80, 0x8F),
    ]


class BasicMultilingualPlane(CharRange):
    """Unicode set for the Basic Multilingual Plane."""

    _ranges: UnicodeRangeList = [
        (0x0020, 0xFFFF),
    ]


EMOJI = list(emoji.EMOJI_DATA.keys())

NON_US_ASCII = NonUsAscii.all() + EMOJI

# Characters that should be encoded in quotes
UNSAFE_CHAR_RE = re.compile(r"[,:;]")


class SafeChar(CharRange):
    """Any character except CONTROL, DQUOTE, ";", ":", ","."""

    _ranges: UnicodeRangeList = [
        (0x21,),  # Control charts before 0x21. 0x22 is "
        (0x23, 0x2B),  # 0x2C is ,
        (0x2D, 0x39),  # 0x3A is : and 0x3B is ;
        (0x3C, 0x7E),  # 0x7E is DEL (control)
    ]


SAFE_CHAR = "".join(WSP + SafeChar.all() + NON_US_ASCII)


class ValueChar(CharRange):
    """Any textual character."""

    _ranges: UnicodeRangeList = [
        (0x21, 0x7E),
    ]


VALUE_CHAR = "".join(
    WSP + ValueChar.all() + NON_US_ASCII + BasicMultilingualPlane.all()
)



"""Tests for the timeline library."""

import datetime
import random
import zoneinfo
from typing import Any

import pytest

from ical.calendar import Calendar
from ical.event import Event
from ical.types.recur import Recur

TZ = zoneinfo.ZoneInfo("America/Regina")


@pytest.fixture(name="calendar")
def fake_calendar(num_events: int, num_instances: int) -> Calendar:
    """Fixture for creating a fake calendar of items."""
    cal = Calendar()
    for i in range(num_events):
        delta = datetime.timedelta(days=int(365 * random.random()))
        cal.events.append(
            Event(
                summary=f"Event {i}",
                start=datetime.date(2022, 2, 1) + delta,
                end=datetime.date(2000, 2, 2) + delta,
                rrule=Recur.from_rrule(f"FREQ=DAILY;COUNT={num_instances}"),
            )
        )
    return cal


@pytest.mark.parametrize(
    "num_events,num_instances",
    [
        (10, 10),
        (10, 100),
        (10, 1000),
        (100, 10),
        (100, 100),
    ],
)
@pytest.mark.benchmark(min_rounds=1, cprofile=True, warmup=False)
def test_benchmark_merged_iter(
    calendar: Calendar, num_events: int, num_instances: int, benchmark: Any
) -> None:
    """Add a benchmark for the merged iterator."""

    def exhaust() -> int:
        nonlocal calendar
        return sum(1 for _ in calendar.timeline_tz(TZ))

    result = benchmark(exhaust)
    assert result == num_events * num_instances
